import * as sqlite3 from 'sqlite3';

type UserEventModel = {
	user: string;
	class: string;
	property: string;
	action: string;
	timestamp: string;
};

type UserStatisticsModel = {
	user: string;
	created: number;
	updated: number;
	removed: number;
	total: number;
};

class DBEvents {
	private static db: sqlite3.Database;

	public static init(path: string): void {
		const db = DBEvents.db = new sqlite3.Database(path);

		db.run(`CREATE TABLE IF NOT EXISTS events (
				id INTEGER PRIMARY KEY AUTOINCREMENT,
				user TEXT,
				class TEXT,
				property TEXT,
				action TEXT,
				timestamp TEXT
		)`, (err) => {
			if(err)
				console.error(err);
		});
	}

	public static async createEvent(user: string, cls: string, property: string, action: string, timestamp: string): Promise<void> {
		const sql = 'INSERT INTO events (user, class, property, action, timestamp) VALUES (?, ?, ?, ?, ?)';
		const params = [user, cls, property, action, timestamp];

		return new Promise<void>(
			(resolve, reject) => {
				DBEvents.db.run(sql, params, function(err) {
					if (err) {
						reject(err);
					} else {
						resolve();
					}
				});
			}
		);
	}

	public static async getStatisticsCount(): Promise<number> {
		const sql = 'SELECT COUNT(*) AS count FROM (SELECT DISTINCT user FROM events);';

		return new Promise<number>((resolve, reject) => {
			DBEvents.db.get<{ count: number }>(sql, [], (err, row) => {
				if (err) {
					reject(err);
				} else {
					resolve(row.count);
				}
			});
		});
	}

	public static async getStatistics(page: number, pageSize: number): Promise<UserStatisticsModel[]> {
		const offset = (page - 1) * pageSize;
		const sql = `
			SELECT
					user,
					SUM(CASE WHEN action = 'create' THEN 1 ELSE 0 END) AS created,
					SUM(CASE WHEN action = 'update' THEN 1 ELSE 0 END) AS updated,
					SUM(CASE WHEN action = 'remove' THEN 1 ELSE 0 END) AS removed,
					(SUM(CASE WHEN action = 'create' THEN 1 ELSE 0 END) + SUM(CASE WHEN action = 'update' THEN 1 ELSE 0 END) + SUM(CASE WHEN action = 'remove' THEN 1 ELSE 0 END)) AS total
			FROM
					events
			GROUP BY
					user
			ORDER BY
					total DESC
			LIMIT ? OFFSET ?;
		`;

		return new Promise<UserStatisticsModel[]>((resolve, reject) => {
			DBEvents.db.all<UserStatisticsModel>(sql, [pageSize, offset], (err, rows) => {
				if(err) {
					reject(err);
				} else {
					resolve(rows);
				}
			});
		});
	}

	public static async getUserEventsCount(user: string): Promise<number> {
		const sql = 'SELECT COUNT(*) AS count FROM events WHERE user = ?;';

		return new Promise<number>((resolve, reject) => {
			DBEvents.db.get<{ count: number }>(sql, [user], (err, row) => {
				if (err) {
					reject(err);
				} else {
					resolve(row.count);
				}
			});
		});
	}

	public static async getUserEvents(user: string, page: number, pageSize: number): Promise<UserEventModel[]> {
		const offset = (page - 1) * pageSize;
		const sql = `
			SELECT user, class, property, action, timestamp
			FROM events
			WHERE user = ?
			ORDER BY timestamp DESC
			LIMIT ? OFFSET ?;
		`;
		return new Promise<UserEventModel[]>((resolve, reject) => {
			DBEvents.db.all<UserEventModel>(sql, [user, pageSize, offset], (err, rows) => {
				if(err) {
					reject(err);
				} else {
					resolve(rows);
				}
			});
		});
	}

	public static async getClassEventsCount(cls: string): Promise<number> {
		const sql = 'SELECT COUNT(*) AS count FROM events WHERE class = ?;';

		return new Promise<number>((resolve, reject) => {
			DBEvents.db.get<{ count: number }>(sql, [cls], (err, row) => {
				if (err) {
					reject(err);
				} else {
					resolve(row.count);
				}
			});
		});
	}

	public static async getClassEvents(cls: string, page: number, pageSize: number): Promise<UserEventModel[]> {
		const offset = (page - 1) * pageSize;
		const sql = `
			SELECT user, class, property, action, timestamp
			FROM events
			WHERE class = ?
			ORDER BY timestamp DESC
			LIMIT ? OFFSET ?;
		`;
		return new Promise<UserEventModel[]>((resolve, reject) => {
			DBEvents.db.all<UserEventModel>(sql, [cls, pageSize, offset], (err, rows) => {
				if(err) {
					reject(err);
				} else {
					resolve(rows);
				}
			});
		});
	}
}

export default DBEvents;